# include <vector>
# include <iostream>
# include <algorithm>

using namespace std;

# include "../library/Structures.h"
# include "../library/ReadPdb.h"
# include "../library/PrintPdb.h"
# include "../library/AtomUtils.h"
# include "../library/MathUtils.h"
# include "../library/rmsd.h"

#include "molecularReplacement.h"

void calcCenter(AtomStruct &Atom1, AtomStruct &Atom2, Vector &ave)
{
	SafeAlloc(ave, 3, "ave");

	ave[X]=(Atom1.x+Atom2.x)*0.5;
	ave[Y]=(Atom1.y+Atom2.y)*0.5;
	ave[Z]=(Atom1.z+Atom2.z)*0.5;
}

void findMostDistantPair(Matrix &aves, int &index1, int &index2)
{
	int Size=aves.size();
	Real dist, max;
	Real dx, dy, dz;

	index1=0;
	index2=1;

	max=0;

	for (int i=0;i<Size;i++)
	{
		for (int j=i+1;j<Size;j++)
		{
			dx=aves[j][X]-aves[i][X];
			dy=aves[j][Y]-aves[i][Y];
			dz=aves[j][Z]-aves[i][Z];

			dist=dx*dx+dy*dy+dz*dz;
			if (dist>=max)
			{
				max=dist;

				index1=i;
				index2=j;
			}
		}
	}
}

void findAverageVector(Matrix &aves, int index1, int index2, VectorStruct &v)
{
	int Size=aves.size();
	int axis;
	Real count=0;
	Vector dx;

	v.x=0;
	v.y=0;
	v.z=0;

	SafeAlloc(dx, 3, "dx");

	dx[X]=abs(aves[index1][X]-aves[index2][X]);
	dx[Y]=abs(aves[index1][Y]-aves[index2][Y]);
	dx[Z]=abs(aves[index1][Z]-aves[index2][Z]);

	if (abs(dx[X])>=abs(dx[Y]) && abs(dx[X])>=abs(dx[Z])) axis=X;
	if (abs(dx[Y])>=abs(dx[X]) && abs(dx[Y])>=abs(dx[Z])) axis=Y;
	if (abs(dx[Z])>=abs(dx[X]) && abs(dx[Z])>=abs(dx[Y])) axis=Z;

	for (int i=0;i<Size;i++)
	{
		for (int j=i+1;j<Size;j++)
		{
			dx[X]=aves[i][X]-aves[j][X];
			dx[Y]=aves[i][Y]-aves[j][Y];
			dx[Z]=aves[i][Z]-aves[j][Z];
			if (dx[axis]<0)
			{
				dx[X]=-dx[X];
				dx[Y]=-dx[Y];
				dx[Z]=-dx[Z];
			}
			v.x+=dx[X];
			v.y+=dx[Y];
			v.z+=dx[Z];
			count+=1.0;
		}
	}
	v.x/=count;
	v.y/=count;
	v.z/=count;

	vectorNormalize(v);

}

Real detectNCS2(vector<ProteinStruct> &Proteins)
{
	int natom;
	int index1, index2;
	Real rmsdNCS;
	Real aveX1, aveY1, aveZ1;
	Real aveX2, aveY2, aveZ2;
	Vector ave;
	Matrix aves;
	VectorStruct v;

	natom=Proteins[0].Atoms.size();

	for (int i=0;i<natom;i++)
	{
		calcCenter(Proteins[0].Atoms[i], Proteins[1].Atoms[i], ave);
		SafePushBack(aves, ave, "aves");
	}
	findMostDistantPair(aves, index1, index2);
	findAverageVector(aves, index1, index2, v);

	calcCenter(Proteins[0].Atoms, aveX1, aveY1, aveZ1);
	calcCenter(Proteins[1].Atoms, aveX2, aveY2, aveZ2);


	ave[X]=(aveX1+aveX2)*0.5;
	ave[Y]=(aveY1+aveY2)*0.5;
	ave[Z]=(aveZ1+aveZ2)*0.5;

	rotateAtomsAroundVector(Proteins[1].Atoms, v, ave[X], ave[Y], ave[Z], pi);
	rmsdNCS=RMSDNoAlign(Proteins[0].Atoms, Proteins[1].Atoms);

	return rmsdNCS;
}

Real detectNCS(vector<ProteinStruct> &Proteins)
{
	int nprot=Proteins.size();
	Real rmsdNCS=UNK_REAL;

	if (nprot==2)
	{
		rmsdNCS=detectNCS2(Proteins);
	}
	return rmsdNCS;
}

Real detectNCS(Vector &degOfFreedom, vector<ProteinStruct> Proteins)
{
	Real rmsdNCS;

	placeProteins(Proteins, degOfFreedom);
	rmsdNCS=detectNCS(Proteins);
	
	return rmsdNCS;	
}
